INTEGRATION_ARTIFACT_BUCKET = "integration_artifact_s3_bucket"
ENVIRONMENT = "environment"
ARTIFACT_MANAGER = "integration_artifact_manager"
CYPRESS_TEST_DIR = "cypress_test_dir"
INTEGRATION_TARGET_URL = "integration_target_url"
TAVERN_ADDITIONAL_ARGS = "tavern_addition_args"
WORKING_TEST_DIR = "verify_environment_new_test_dir"
TAVERN_TEST_DIR = "tavern_test_dir"
DEFAULT_TAVERN_TEST_DIR = "src/integrationtest/tavern"
PROMOTE_ARTIFACT = "promote"
ROLE = "role"
APPLICATION = "application"
APPLICATION_GROUP = "application_group"



from pybuilder.core import task, Project, Logger, depends, after, init
from pybuilder.reactor import Reactor

import pybuilder_integration.tasks
from pybuilder_integration.properties import *


@init
def init_plugin(project):
    project.set_property("tavern_addition_args", [])
    project.set_property_if_unset(CYPRESS_TEST_DIR, "src/integrationtest/cypress")
    project.set_property_if_unset(TAVERN_TEST_DIR, DEFAULT_TAVERN_TEST_DIR)
    project.plugin_depends_on("pytest")
    project.plugin_depends_on("tavern")


@task(description="Runs integration tests against a CI/Prod environment."
                  "\t1. Run current build integration tests found in ${dir_dist}\n"
                  f"\t2. Run integration tests found in 'LATEST-{ENVIRONMENT}' managed by ${ARTIFACT_MANAGER}\n"
                  f"\t3. Promote current build integration tests to 'LATEST-{ENVIRONMENT}' (disable with ${PROMOTE_ARTIFACT})\n"
                  f"\t${INTEGRATION_TARGET_URL} - (required) Full URL target for tests\n"
                  f"\t${ENVIRONMENT} - (required) Environment that is being tested (ci/prod)\n"
                  f"\t${PROMOTE_ARTIFACT} - Promote integration tests to LATEST-${ENVIRONMENT} (default TRUE)\n"
      )
def verify_environment(project: Project, logger: Logger, reactor: Reactor):
    tasks.verify_environment(project, logger, reactor)


@task(description="Run integration tests using a cypress spec. Requires NPM installed.\n"
                  f"\t{INTEGRATION_TARGET_URL} - (required) Full URL target for cypress tests\n"
                  f"\t{CYPRESS_TEST_DIR} - directory for test specification (src/integrationtest/cypress)\n"
      )
def verify_cypress(project: Project, logger: Logger, reactor: Reactor):
    tasks.verify_cypress(project, logger, reactor)


@task(description="Run integration tests using tavern specifications.\n"
                  f"\t{TAVERN_TEST_DIR} - directory containing tavern specifications ({DEFAULT_TAVERN_TEST_DIR})")
def verify_tavern(project: Project, logger: Logger, reactor: Reactor):
    tasks.verify_tavern(project, logger, reactor)


@task(description="Run verify_tavern and verify_cypress")
@depends("publish", "verify_tavern", "verify_cypress")
def verify_package():
    pass


@task(description="Package artifacts for publishing in integration tests")
def package_artifacts(project: Project):
    package_tavern_artifacts(project)
    package_cypress_artifacts(project)


@task(description="Package tavern artifacts for publishing in integration tests")
def package_tavern_artifacts(project: Project):
    test_dir = project.expand_path(f"${TAVERN_TEST_DIR}")
    tasks.package_artifacts(project, test_dir, "tavern")


@task(description="Package cypress artifacts for publishing in integration tests")
def package_cypress_artifacts(project: Project):
    test_dir = project.expand_path(f"${CYPRESS_TEST_DIR}")
    tasks.package_artifacts(project, test_dir, "cypress")



import os
import shutil
from typing import Dict

from pybuilder.core import Project, Logger
from pybuilder.errors import BuildFailedException
from pybuilder.reactor import Reactor

from pybuilder_integration import exec_utility
from pybuilder_integration.directory_utility import get_latest_distribution_directory, \
    get_latest_zipped_distribution_directory
from pybuilder_integration.properties import INTEGRATION_ARTIFACT_BUCKET, ENVIRONMENT, ROLE, APPLICATION, \
    APPLICATION_GROUP, ARTIFACT_MANAGER


class ArtifactManager:
    def __init__(self, name, identifier):
        self.identifier = identifier
        self.friendly_name = name

    def upload(self, file: str, project: Project, logger: Logger, reactor: Reactor):
        pass

    def download_artifacts(self, project: Project, logger: Logger, reactor: Reactor):
        pass


class S3ArtifactManager(ArtifactManager):

    def __init__(self):
        super().__init__("AWS S3 Artifact Manager", "S3")

    def upload(self, file: str, project: Project, logger: Logger, reactor: Reactor):
        if project.get_property("abort_upload","false") != "false":
            return
        # First make sure bucket exists
        self.create_bucket(logger, project, reactor)
        relative_path = get_latest_artifact_destination(logger, project)
        self._s3_transfer(file, relative_path, project, reactor, logger, recursive=False)
        relative_path = get_versioned_artifact_destination(logger, project)
        self._s3_transfer(file, relative_path, project, reactor, logger, recursive=False)

    def download_artifacts(self, project: Project, logger: Logger, reactor: Reactor):
        # this is a noop if there is no bucket
        if not self.does_bucket_exist(logger, project, reactor):
            return
        s3_location = get_latest_artifact_destination(logger, project)
        zipped_directory = get_latest_zipped_distribution_directory(project)
        self._s3_transfer(source=s3_location,
                          destination=zipped_directory,
                          project=project,
                          logger=logger,
                          reactor=reactor)
        return _unzip_downloaded_artifacts(zipped_directory, get_latest_distribution_directory(project), logger)

    @staticmethod
    def _s3_transfer(source, destination, project, reactor, logger, recursive=True):
        logger.info(f"Proceeding to transfer {source} to {destination}")
        S3ArtifactManager.verify_aws_cli(reactor)
        #  aws s3 cp myDir s3://mybucket/ --recursive
        args = [
            's3',
            'cp',
            source,
            destination
        ]
        if recursive:
            args.append("--recursive")
        exec_utility.exec_command(command_name='aws',
                                  args=args,
                                  failure_message=f"Failed to transfer integration artifacts to {destination}",
                                  log_file_name='s3-artifact-transfer',
                                  project=project,
                                  reactor=reactor,
                                  logger=logger,
                                  report=False)

    @staticmethod
    def verify_aws_cli(reactor):
        reactor.pybuilder_venv.verify_can_execute(command_and_arguments=["aws", "--version"],
                                                  prerequisite="aws cli",
                                                  caller="integration_tests")

    def create_bucket(self, logger, project, reactor):
        app_group, app_name, bucket, environment, role = get_project_metadata(logger, project)
        S3ArtifactManager.verify_aws_cli(reactor)
        res = self.does_bucket_exist(logger, project, reactor)
        if res:
            return
        exec_utility.exec_command(command_name='aws',
                                  args=[
                                      's3api',
                                      'create-bucket',
                                      '--acl',
                                      'private',
                                      '--bucket',
                                      bucket
                                  ],
                                  failure_message=f"Failed to create bucket",
                                  log_file_name='s3-create-bucket',
                                  project=project,
                                  reactor=reactor,
                                  logger=logger,
                                  report=False)

    def does_bucket_exist(self, logger, project, reactor):
        app_group, app_name, bucket, environment, role = get_project_metadata(logger, project)
        return exec_utility.exec_command(command_name='aws',
                                         args=[
                                             's3api',
                                             'head-bucket',
                                             '--bucket',
                                             bucket
                                         ],
                                         failure_message=f"Failed to find bucket",
                                         log_file_name='s3-head-bucket',
                                         project=project,
                                         reactor=reactor,
                                         logger=logger,
                                         raise_exception=False,
                                         report=False)


artifact_managers: Dict[str, S3ArtifactManager] = {}
manager = S3ArtifactManager()
artifact_managers[manager.identifier] = manager


def get_artifact_manager(project: Project) -> ArtifactManager:
    manager_id = project.get_property(ARTIFACT_MANAGER, "S3")
    global artifact_managers
    manager = artifact_managers.get(manager_id)
    if not manager:
        raise BuildFailedException(f"Failed to find appropriate artifact manager for {manager_id}")
    return manager


def extract_application_role(logger, project):
    # expect a fully populated naming scheme
    app_group = project.get_property(APPLICATION_GROUP)
    app_name = project.get_property(APPLICATION)
    role = project.get_property(ROLE)
    #  if we do not have role set, assume it is all baked into the project name
    if not role:
        project_name = project.name
        if project_name.find("-") >= 0:
            split = project_name.split("-")
            if len(split) >= 3:
                app_group = split[0]
                app_name = split[1]
                if len(split) == 3:
                    role = split[2]
                else:
                    role = "-".join(split[2:])
            else:
                logger.info(
                    f"Unexpected naming format expected <Application Group>-<Application>-<Role> got {project_name}")
        else:
            app_name = project_name
            role = 'Unknown'
            logger.info(
                f"Unexpected naming format expected <Application Group>-<Application>-<Role> got {project_name}")
    if not app_group:
        if app_name and app_name.find("-") >= 0:
            # backwards compat for the simple days before app groups
            split = app_name.split("-")
            app_group = split[0]
            app_name = split[1]
        else:
            app_group = "Unknown"
    return app_group, app_name, role


def _unzip_downloaded_artifacts(dir_with_zips: str, destination: str, logger: Logger) -> str:
    for file in os.listdir(dir_with_zips):
        # expect {tool}-{self.project.name}.zip
        if os.path.basename(file).find("tavern") >= 0:
            shutil.unpack_archive(filename=os.path.join(dir_with_zips, file), extract_dir=f"{destination}/tavern",
                                  format="zip")
        elif os.path.basename(file).find("cypress") >= 0:
            shutil.unpack_archive(filename=os.path.join(dir_with_zips, file), extract_dir=f"{destination}/cypress",
                                  format="zip")
        else:
            logger.warn(f"Unexpected file name in downloaded artifacts {file}")
    return destination


def get_project_metadata(logger: Logger, project: Project):
    app_group, app_name, role = extract_application_role(logger, project)
    environment = project.get_mandatory_property(ENVIRONMENT)
    bucket = project.get_property(INTEGRATION_ARTIFACT_BUCKET, f"integration-artifacts-{app_group}-{app_name}")
    return app_group, app_name, bucket, environment, role


def get_versioned_artifact_destination(logger, project):
    app_group, app_name, bucket, environment, role = get_project_metadata(logger, project)
    return f"s3://{bucket}/{role}/{project.version}/"


def get_latest_artifact_destination(logger, project):
    app_group, app_name, bucket, environment, role = get_project_metadata(logger, project)
    return f"s3://{bucket}/LATEST-{environment}/"



from pybuilder.errors import BuildFailedException
from pybuilder.pluginhelper.external_command import ExternalCommandBuilder, ExternalCommandResult
from pybuilder.utils import read_file

from pybuilder_integration.directory_utility import prepare_logs_directory, prepare_reports_directory


def exec_command(command_name,
                 args,
                 failure_message,
                 log_file_name,
                 project,
                 reactor,
                 logger,
                 working_dir=None,
                 raise_exception=True,
                 report=True,
                 env_vars=None):
    if working_dir:
        command = WorkingDirCommandBuilder(command_name=command_name,
                                           project=project,
                                           reactor=reactor,
                                           env=env_vars,
                                           cwd=working_dir)
    else:
        command = ExternalCommandBuilder(command_name=command_name,
                                         project=project,
                                         reactor=reactor)

    for arg in args:
        command.use_argument(arg)
    if report:
        directory = prepare_reports_directory(project)
    else:
        directory = prepare_logs_directory(project)
    outfile_name = "{}/{}".format(directory, log_file_name)
    res = command.run(outfile_name)
    if res.exit_code != 0:
        if raise_exception:
            raise BuildFailedException(failure_message)
        else:
            logger.warn(failure_message)
            return False
    return True


class WorkingDirCommandBuilder(ExternalCommandBuilder):

    def __init__(self, command_name, project, cwd, reactor, env=None):
        super(WorkingDirCommandBuilder, self).__init__(command_name, project, reactor)
        self.env = env if env else {}
        self.cwd = cwd

    def run(self, outfile_name):
        error_file_name = "{0}.err".format(outfile_name)
        return_code = self._env.execute_command(self.parts, outfile_name, env=self.env, cwd=self.cwd)
        error_file_lines = read_file(error_file_name)
        outfile_lines = read_file(outfile_name)

        return ExternalCommandResult(return_code,
                                     outfile_name, outfile_lines,
                                     error_file_name, error_file_lines)



import os
import shutil


def prepare_reports_directory(project):
    return prepare_directory("$dir_reports", project)


def prepare_logs_directory(project):
    return prepare_directory("$dir_logs", project)


def prepare_dist_directory(project):
    return prepare_directory("$dir_dist", project)


def prepare_directory(dir_variable, project):
    package__format = f"{dir_variable}/integration"
    reports_dir = project.expand_path(package__format)
    return _ensure_directory_exists(reports_dir)


def _ensure_directory_exists(path):
    if not os.path.exists(path):
        os.makedirs(path)
    return path


def get_working_distribution_directory(project):
    dist_directory = prepare_dist_directory(project)
    return _ensure_directory_exists(f"{dist_directory}/working")


def get_latest_distribution_directory(project):
    dist_directory = prepare_dist_directory(project)
    return _ensure_directory_exists(f"{dist_directory}/LATEST")


def get_latest_zipped_distribution_directory(project):
    dist_directory = prepare_dist_directory(project)
    return _ensure_directory_exists(f"{dist_directory}/LATEST/zipped")


def get_local_zip_artifact_path(tool, project, include_ending=False):
    artifact = f"{prepare_dist_directory(project)}/{tool}-{project.name}"
    if include_ending:
        return f"{artifact}.zip"
    return artifact


def package_artifacts(project, test_dir, tool):
    # Make a copy for easy access in environment validation
    working_dir = get_working_distribution_directory(project)
    shutil.copytree(test_dir, f"{working_dir}/{tool}",dirs_exist_ok=True)
    # package a copy for distribution
    # zip up the test and add them to the integration test dist directory
    base_name = get_local_zip_artifact_path(tool=tool, project=project)
    shutil.make_archive(base_name=base_name,
                        format="zip",
                        root_dir=test_dir)



from pybuilder.core import Logger, Project
from pybuilder.reactor import Reactor

from pybuilder_integration.exec_utility import exec_command


def install_cypress(logger: Logger, project: Project, reactor: Reactor):
    _install_npm_tool(tool_name="cypress", logger=logger, project=project, reactor=reactor)
    _install_npm_tool(tool_name="typescript", logger=logger, project=project, reactor=reactor)


def _install_npm_tool(tool_name: str, logger: Logger, project: Project, reactor: Reactor):
    reactor.pybuilder_venv.verify_can_execute(
        command_and_arguments=["npm", "--version"], prerequisite="npm", caller="integration_tests")
    logger.info(f"Ensuring {tool_name} is installed")
    exec_command('npm', ['install', tool_name], f'Failed to install {tool_name} - required for integration tests',
                 f'{tool_name}_npm_install', project, reactor, logger,report=False)



#!/usr/bin/env python
#   -*- coding: utf-8 -*-

from setuptools import setup
from setuptools.command.install import install as _install

class install(_install):
    def pre_install_script(self):
        pass

    def post_install_script(self):
        pass

    def run(self):
        self.pre_install_script()

        _install.run(self)

        self.post_install_script()

if __name__ == '__main__':
    setup(
        name = 'pybuilder-integration',
        version = '49',
        description = 'A pybuilder plugin that runs integration tests (Tavern & Cypress) against a target.',
        long_description = 'A pybuilder plugin that runs integration tests against a target.  This is intended to be a broader scope than unit-tests encompassing dependant functionality.',
        long_description_content_type = None,
        classifiers = [
            'Development Status :: 3 - Alpha',
            'Programming Language :: Python'
        ],
        keywords = '',

        author = '',
        author_email = '',
        maintainer = '',
        maintainer_email = '',

        license = 'Apache 2.0',

        url = 'https://github.com/rspitler/pybuilder-integration',
        project_urls = {},

        scripts = [],
        packages = ['pybuilder_integration'],
        namespace_packages = [],
        py_modules = [],
        entry_points = {},
        data_files = [],
        package_data = {},
        install_requires = [
            'pytest',
            'tavern'
        ],
        dependency_links = [],
        zip_safe = True,
        cmdclass = {'install': install},
        python_requires = '',
        obsoletes = [],
    )



import os

import pytest
from pybuilder.core import Project, Logger, init, RequirementsFile
from pybuilder.errors import BuildFailedException
from pybuilder.install_utils import install_dependencies
from pybuilder.reactor import Reactor

from pybuilder_integration import exec_utility
from pybuilder_integration.artifact_manager import get_artifact_manager
from pybuilder_integration.directory_utility import prepare_dist_directory, get_working_distribution_directory, \
    package_artifacts, prepare_reports_directory, get_local_zip_artifact_path
from pybuilder_integration.properties import *
from pybuilder_integration.tool_utility import install_cypress


def integration_artifact_push(project: Project, logger: Logger, reactor: Reactor):
    logger.info("Starting upload of integration artifacts")
    manager = get_artifact_manager(project)
    for tool in ["tavern", "cypress"]:
        artifact_file = get_local_zip_artifact_path(tool=tool, project=project, include_ending=True)
        if os.path.exists(artifact_file):
            logger.info(f"Starting upload of integration artifact: {os.path.basename(artifact_file)} to: {manager.friendly_name}")
            manager.upload(file=artifact_file, project=project, logger=logger, reactor=reactor)


def verify_environment(project: Project, logger: Logger, reactor: Reactor):
    dist_directory = project.get_property(WORKING_TEST_DIR, get_working_distribution_directory(project))
    logger.info(f"Preparing to run tests found in: {dist_directory}")
    _run_tests_in_directory(dist_directory, logger, project, reactor)
    artifact_manager = get_artifact_manager(project=project)
    latest_directory = artifact_manager.download_artifacts(project=project, logger=logger, reactor=reactor)
    _run_tests_in_directory(latest_directory, logger, project, reactor)
    if project.get_property(PROMOTE_ARTIFACT, True):
        integration_artifact_push(project=project, logger=logger, reactor=reactor)


def _run_tests_in_directory(dist_directory, logger, project, reactor):
    cypress_test_path = f"{dist_directory}/cypress"
    if os.path.exists(cypress_test_path):
        logger.info(f"Found cypress tests - starting run")
        _run_cypress_tests_in_directory(work_dir=cypress_test_path,
                                        logger=logger,
                                        project=project,
                                        reactor=reactor)
    tavern_test_path = f"{dist_directory}/tavern"
    if os.path.exists(tavern_test_path):
        logger.info(f"Found tavern tests - starting run")
        _run_tavern_tests_in_dir(test_dir=tavern_test_path,
                                 logger=logger,
                                 project=project,
                                 reactor=reactor)


def verify_cypress(project: Project, logger: Logger, reactor: Reactor):
    # Get directories with test and cypress executable
    work_dir = project.expand_path(f"${CYPRESS_TEST_DIR}")
    if _run_cypress_tests_in_directory(work_dir=work_dir, logger=logger, project=project,
                                       reactor=reactor):
        package_artifacts(project, work_dir, "cypress")


def _run_cypress_tests_in_directory(work_dir, logger, project, reactor: Reactor):
    target_url = project.get_mandatory_property(INTEGRATION_TARGET_URL)
    environment = project.get_mandatory_property(ENVIRONMENT)
    if not os.path.exists(work_dir):
        logger.info("Skipping cypress run: no tests")
        return False
    logger.info(f"Found {len(os.listdir(work_dir))} files in cypress test directory")
    # Validate NPM install and Install cypress
    install_cypress(project=project, logger=logger, reactor=reactor)
    executable = project.expand_path("./node_modules/cypress/bin/cypress")
    results_file, run_name = get_test_report_file(project=project,test_dir=work_dir,tool="cypress")
    # Run the actual tests against the baseURL provided by ${integration_target}
    args = ["run", "--config", f"baseUrl={target_url}", "--reporter-options",
            f"mochaFile={results_file}"]
    config_file_path = f'{environment}-config.json'
    if os.path.exists(os.path.join(work_dir, config_file_path)):
        args.append("--config-file")
        args.append(config_file_path)
    logger.info(f"Running cypress on host: {target_url}")
    exec_utility.exec_command(command_name=executable, args=args,
                              failure_message="Failed to execute cypress tests", log_file_name='cypress_run.log',
                              project=project, reactor=reactor, logger=logger, working_dir=work_dir, report=False)
    return True


def verify_tavern(project: Project, logger: Logger, reactor: Reactor):
    # Expand the directory to get full path
    test_dir = project.expand_path(f"${TAVERN_TEST_DIR}")
    # Run the tests in the directory
    if _run_tavern_tests_in_dir(test_dir, logger, project, reactor):
        package_artifacts(project, test_dir, "tavern")


def _run_tavern_tests_in_dir(test_dir: str, logger: Logger, project: Project, reactor: Reactor):
    logger.info("Running tavern tests: {}".format(test_dir))
    if not os.path.exists(test_dir):
        logger.info("Skipping tavern run: no tests")
        return False
    logger.info(f"Found {len(os.listdir(test_dir))} files in tavern test directory")
    # todo is this unique enough for each run?
    output_file, run_name = get_test_report_file(project, test_dir)
    from sys import path as syspath
    syspath.insert(0, test_dir)
    # install any requirements that my exist
    requirements_file = os.path.join(test_dir, "requirements.txt")
    if os.path.exists(requirements_file):
        dependency = RequirementsFile(requirements_file)
        install_dependencies(logger, project, dependency, reactor.pybuilder_venv, "install_batch")
    extra_args = [project.expand(prop) for prop in project.get_property(TAVERN_ADDITIONAL_ARGS, [])]
    args = ["--junit-xml", f"{output_file}", test_dir] + extra_args
    if project.get_property("verbose"):
        args.append("-s")
        args.append("-v")
    os.environ['TARGET'] = project.get_property(INTEGRATION_TARGET_URL)
    ret = pytest.main(args)
    if ret != 0:
        raise BuildFailedException(f"Tavern tests failed see complete output here - {output_file}")
    return True


def get_test_report_file(project, test_dir,tool="tavern"):
    run_name = os.path.basename(os.path.realpath(os.path.join(test_dir, os.pardir)))
    output_file = os.path.join(prepare_reports_directory(project), f"{tool}-{run_name}.out.xml")
    return output_file, run_name
